from __future__ import annotations

from typing import Dict

from ..data import Compound, Reaction

__all__ = [
    "MALFORMED_LINE_STARTS",
    "_rename",
    "_set_gibbs0",
    "_set_name",
    "_add_database_link",
    "_add_type",
]

# Often lines starting with these identifiers are malformed
MALFORMED_LINE_STARTS = {
    "/",
    "COMMENT",
    "CITATIONS",
    "^CITATIONS",
    "SYNONYMS",
    "#",
}


def _rename(content: str) -> str:
    """Remove garbage from compound and reaction ids."""
    return (
        content.replace("<i>", "")
        .replace("</i>", "")
        .replace("<SUP>", "")
        .replace("</SUP>", "")
        .replace("<sup>", "")
        .replace("</sup>", "")
        .replace("<sub>", "")
        .replace("</sub>", "")
        .replace("<SUB>", "")
        .replace("</SUB>", "")
        .replace("&", "")
        .replace(";", "")
        .replace("|", "")
    )


def _set_gibbs0(dictionary: Dict[str, Reaction | Compound], id_: str, gibbs0: str) -> None:
    try:
        dictionary[id_].gibbs0 = float(gibbs0)
    except ValueError:  # conversion failed
        pass


def _set_name(dictionary: Dict[str, Reaction | Compound], id_: str, name: str) -> None:
    dictionary[id_].name = _rename(name)


def _add_database_link(
    dictionary: Dict[str, Reaction | Compound],
    id_: str,
    content: str,
    db_to_resource: dict[str, str],
) -> None:
    """Short description.

    Database links are of form DBLINKS - (REFMET "Tryptophan" NIL |midford| 3697479617 NIL NIL)
    so content will be (REFMET "Tryptophan" NIL |midford| 3697479617 NIL NIL)
    """
    database, database_id, *_ = content[1:-1].split(" ")
    if (db := db_to_resource.get(database, None)) is not None:
        dictionary[id_].database_links.setdefault(db, set()).add(database_id[1:-1])


def _add_type(
    dictionary: Dict[str, Reaction | Compound],
    id_: str,
    type_: str,
    type_map: Dict[str, str],
) -> None:
    """Short description."""
    dictionary[id_].types.append(type_map.get(type_, type_))



from __future__ import annotations

from pathlib import Path
from typing import Dict, Set

from .data import (
    Compound,
    Cyc,
    Enzyme,
    KineticData,
    Monomer,
    MonomerAnnotation,
    Reaction,
    Result,
)
from .defaults import COMPARTMENT_MAP, COMPARTMENT_SUFFIXES, MANUAL_ADDITIONS, TYPE_MAP
from .parsing import parse_pgdb
from .repairing import (
    fix_add_important_compounds,
    fix_create_compartment_variants,
    fix_create_reaction_variants,
    fix_filter_garbage_reactions,
    fix_set_reaction_stoichiometry,
    fix_unify_reaction_direction,
    get_gpr_associations,
    get_highest_compound_type,
    get_kinetic_data,
)

__all__ = [
    "repair",
    "parse_and_repair_pgdb",
    "Compound",
    "Cyc",
    "Enzyme",
    "KineticData",
    "Monomer",
    "MonomerAnnotation",
    "Reaction",
    "COMPARTMENT_MAP",
    "COMPARTMENT_SUFFIXES",
    "MANUAL_ADDITIONS",
    "TYPE_MAP",
]


def combine_sequence_and_annotation(
    sequences: Dict[str, str], annotation: Dict[str, MonomerAnnotation]
) -> Dict[str, Monomer]:
    return {
        k: Monomer(id=k, sequence=sequences[k], database_links=v.database_links)
        for k, v in annotation.items()
        if k in sequences
    }


def repair(
    parse_result: Result,
    compartment_map: Dict[str, str],
    manual_additions: Dict[str, Compound],
    compartment_suffixes: Dict[str, str],
) -> Cyc:
    compounds = parse_result.compounds
    reactions = parse_result.reactions
    monomer_annotation = parse_result.monomer_annotation
    complexes = parse_result.complexes
    enzymes = parse_result.enzymes
    sequences = parse_result.sequences

    monomers = combine_sequence_and_annotation(sequences, monomer_annotation)

    # Generate dervied data
    compound_types = get_highest_compound_type(compounds)
    gpr_annotations = get_gpr_associations(reactions, enzymes, complexes)
    kinetic_data = get_kinetic_data(reactions, enzymes)

    # Small fixes
    compounds = fix_add_important_compounds(compounds, manual_additions)
    reactions = fix_unify_reaction_direction(reactions)

    # Larger fixes
    reactions = fix_create_reaction_variants(reactions, compounds, compound_types)
    reactions = fix_filter_garbage_reactions(reactions, compounds)
    reactions = fix_create_compartment_variants(
        reactions,
        compounds,
        compartment_map,
        compartment_suffixes,
    )
    reactions = fix_set_reaction_stoichiometry(reactions)

    # Post fixing
    used_compartments: Set[str] = {i.compartment for i in compounds.values() if i.compartment is not None}
    compartments: Dict[str, str] = {i: compartment_suffixes[i] for i in used_compartments}
    return Cyc(compounds, reactions, compartments, gpr_annotations, kinetic_data, monomers)


def parse_and_repair_pgdb(
    pgdb_path: Path,
    compartment_map: Dict[str, str] | None = None,
    type_map: Dict[str, str] | None = None,
    manual_additions: Dict[str, Compound] | None = None,
    compartment_suffixes: Dict[str, str] | None = None,
) -> Cyc:
    if compartment_map is None:
        compartment_map = COMPARTMENT_MAP
    if type_map is None:
        type_map = TYPE_MAP
    if manual_additions is None:
        manual_additions = MANUAL_ADDITIONS
    if compartment_suffixes is None:
        compartment_suffixes = COMPARTMENT_SUFFIXES
    return repair(
        parse_pgdb(pgdb_path, type_map),
        compartment_map,
        manual_additions,
        compartment_suffixes,
    )



from __future__ import annotations

import logging
from functools import partial
from typing import Any, Callable, Dict, Iterable

from ..data import MonomerAnnotation
from .shared import MALFORMED_LINE_STARTS, _add_database_link

__all__ = ["parse_monomer_annotation"]

logger = logging.getLogger(__name__)
Action = Callable[[Any, Any, Any], None]


def _set_gene_product(genes: Dict[str, MonomerAnnotation], id_: str, product: str) -> None:
    genes[id_].product = product


def parse_monomer_annotation(
    file: Iterable[str],
    actions: Dict[str, Action] | None = None,
    db_to_resource: dict[str, str] | None = None,
) -> Dict[str, MonomerAnnotation]:
    if db_to_resource is None:
        db_to_resource = {
            "ARRAYEXPRESS": "arrayexpress",
            "ASAP": "asap",
            "CGD": "cgd",
            "CGSC": "cgsc",
            "ECHOBASE": "echobase",
            "ENSEMBL": "ensembl",
            "ENSEMBLGENOMES-GN": "ensembl",
            "ENSEMBLGENOMES-TR": "ensembl",
            "GENECARDS": "genecards",
            "GO": "GO",
            "GOA": "goa",
            "GRAMENE": "gramene.gene",
            "IMG": "img.gene",
            "INTERPRO": "interpro",
            "KEGG": "kegg.genes",
            "MAIZEGDB": "maizegdb.locus",
            "MGI": "MGI",
            "MIM": "mim",
            "NCBI-GENE": "ncbigene",
            "PDB": "pdb",
            "PHYTOZOME": "phytozome.locus",
            "PID": "pid.pathway",
            "REFSEQ": "refseq",
            "SGD": "sgb",
            "SGN": "sgn",
            "STRING": "string",
            "SUBTILIST": "subtilist",
            "SUBTIWIKI": "subtiwiki",
            "TAIR": "tair.gene",
            "UNIGENE": "unigene",
            "UNIPROT": "uniprot",
        }
    if actions is None:
        actions = {
            "DBLINKS": partial(_add_database_link, db_to_resource=db_to_resource),
            "PRODUCT": _set_gene_product,
        }

    genes: Dict[str, MonomerAnnotation] = {}
    id_ = ""
    for line in file:
        if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
            continue
        try:
            identifier, content = line.rstrip().split(" - ", maxsplit=1)
        except ValueError:
            logger.info(f"Malformed line in genes.dat: {line}")
            continue

        if identifier == "UNIQUE-ID":
            id_ = content
            genes[content] = MonomerAnnotation(id=content)
        else:
            if (action := actions.get(identifier, None)) is not None:
                action(genes, id_, content)
    return {
        product: MonomerAnnotation(id=product, database_links=i.database_links)
        for i in genes.values()
        if (product := i.product) is not None
    }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Set, Tuple

__all__ = [
    "Monomer",
    "MonomerAnnotation",
    "Compound",
    "Reaction",
    "Enzyme",
]


@dataclass
class Compound:
    id: str
    base_id: str
    charge: int = 0
    compartment: str = "CYTOSOL"
    smiles: str | None = None
    name: str | None = None
    gibbs0: float | None = None
    types: List[str] = field(default_factory=list)
    formula: Dict[str, float] = field(default_factory=dict)
    database_links: Dict[str, Set[str]] = field(default_factory=dict)


@dataclass
class Reaction:
    id: str
    base_id: str
    bounds: Tuple[float, float] = (0, 0)
    name: str | None = None
    ec: str | None = None
    gibbs0: float | None = None
    direction: str = "LEFT-TO-RIGHT"
    reversible: bool = False
    transmembrane: bool = False
    substrates: Dict[str, float] = field(default_factory=dict)
    substrate_compartments: Dict[str, str] = field(default_factory=dict)
    products: Dict[str, float] = field(default_factory=dict)
    product_compartments: Dict[str, str] = field(default_factory=dict)
    types: List[str] = field(default_factory=list)
    locations: List[str] = field(default_factory=list)
    pathways: Set[str] = field(default_factory=set)
    enzrxns: Set[str] = field(default_factory=set)
    database_links: Dict[str, Set[str]] = field(default_factory=dict)
    _var: int | None = None
    # Later additions
    compartment: str | Tuple[str, str] | None = None
    stoichiometries: Dict[str, float] = field(default_factory=dict)


@dataclass
class MonomerAnnotation:
    id: str
    product: str | None = None  # Later used as id
    database_links: Dict[str, Set[str]] = field(default_factory=dict)


@dataclass
class Monomer:
    id: str
    sequence: str
    database_links: Dict[str, Set[str]] = field(default_factory=dict)


@dataclass
class KineticData:
    kcat: Dict[str, float] = field(default_factory=dict)
    km: Dict[str, float] = field(default_factory=dict)
    vmax: Dict[str, float] = field(default_factory=dict)


@dataclass
class Enzyme:
    id: str
    cplx_or_monomer: str | None = None
    kinetic_data: KineticData = field(default_factory=KineticData)


@dataclass
class Result:
    compounds: Dict[str, Compound]
    reactions: Dict[str, Reaction]
    monomer_annotation: Dict[str, MonomerAnnotation]
    complexes: Dict[str, Set[str]]
    enzymes: Dict[str, Enzyme]
    sequences: Dict[str, str]


@dataclass
class Cyc:
    compounds: Dict[str, Compound]
    reactions: Dict[str, Reaction]
    compartments: Dict[str, str]
    gpr_annotations: Dict[str, List[Set[str]]]  # rxn: [monomer*]
    kinetic_data: Dict[str, Dict[str, KineticData]]  # rxn: enzyme: data
    monomers: Dict[str, Monomer]  # monomer: sequence / annotation



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List

from ..data import Compound


def get_highest_compound_type(compounds: Dict[str, Compound]) -> Dict[str, List[str]]:
    """Return (type: list(cpds)) dictionary.

    Only uses the highest-level type
    """
    types = defaultdict(list)
    for id_, cpd in compounds.items():
        if bool(cpd.types):
            # Only use highest level
            types[cpd.types[-1] + "_c"].append(id_)
    return dict(types)



from __future__ import annotations

import logging
from typing import Any, Callable, Dict, Iterable

from ..data import Enzyme
from .shared import MALFORMED_LINE_STARTS

__all__ = ["parse_enzymes"]

logger = logging.getLogger(__name__)
Action = Callable[[Any, Any, Any], None]
SubAction = Dict[str, Callable[[Any, Any, Any, Any], None]]


def _set_cplx_or_monomer(enzrxns: Dict[str, Enzyme], id_: str, cplx_or_monomer: str) -> None:
    enzrxns[id_].cplx_or_monomer = cplx_or_monomer


def _add_kcat(enzrxns: Dict[str, Enzyme], id_: str, substrate: str, kcat: str) -> None:
    try:
        enzrxns[id_].kinetic_data.kcat.setdefault(substrate, float(kcat))
    except ValueError:  # conversion failed
        pass


def _add_km(enzrxns: Dict[str, Enzyme], id_: str, substrate: str, km: str) -> None:
    try:
        enzrxns[id_].kinetic_data.km.setdefault(substrate, float(km))
    except ValueError:  # conversion failed
        pass


def _add_vmax(enzrxns: Dict[str, Enzyme], id_: str, substrate: str, vmax: str) -> None:
    try:
        enzrxns[id_].kinetic_data.vmax.setdefault(substrate, float(vmax))
    except ValueError:  # conversion failed
        pass


def parse_enzymes(
    file: Iterable[str],
    actions: Dict[str, Action] | None = None,
    sub_actions: Dict[str, SubAction] | None = None,
) -> Dict[str, Enzyme]:

    if actions is None:
        actions = {
            "ENZYME": _set_cplx_or_monomer,
        }
    if sub_actions is None:
        sub_actions = {
            "^SUBSTRATE": {"KM": _add_km, "VMAX": _add_vmax, "KCAT": _add_kcat},
        }

    id_ = ""
    enzrxns = {}
    last_identifier = ""
    last_content = ""
    for line in file:
        if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
            continue
        try:
            identifier, content = line.rstrip().split(" - ", maxsplit=1)
        except ValueError:
            logger.info(f"Malformed line in enzymes.dat {line}")
            continue

        if identifier == "UNIQUE-ID":
            id_ = content
            enzrxns[id_] = Enzyme(id=id_)
        elif identifier.startswith("^"):
            if (subaction := sub_actions.get(identifier, {}).get(last_identifier, None)) is not None:
                subaction(enzrxns, id_, content, last_content)
        else:
            if (action := actions.get(identifier, None)) is not None:
                action(enzrxns, id_, content)
                last_identifier = identifier
                last_content = content
    return enzrxns



from __future__ import annotations

from typing import Dict

from ..data import Enzyme, KineticData, Reaction


def get_kinetic_data(
    reactions: Dict[str, Reaction],
    enzymes: Dict[str, Enzyme],
) -> Dict[str, Dict[str, KineticData]]:
    res: Dict[str, Dict[str, KineticData]] = {}
    for name, reaction in reactions.items():
        for enzyme in reaction.enzrxns:
            if (enz := enzymes.get(enzyme)) is not None:
                res.setdefault(name, {}).setdefault(enzyme, enz.kinetic_data)
    return res



from __future__ import annotations

__all__ = [
    "fix_add_important_compounds",
    "fix_create_compartment_variants",
    "fix_create_reaction_variants",
    "fix_filter_garbage_reactions",
    "fix_set_reaction_stoichiometry",
    "fix_unify_reaction_direction",
    "get_highest_compound_type",
    "get_gpr_associations",
    "get_kinetic_data",
]


from .fix_add_important_cpds import fix_add_important_compounds
from .fix_create_compartment_variants import fix_create_compartment_variants
from .fix_create_reaction_variants import fix_create_reaction_variants
from .fix_filter_garbage import fix_filter_garbage_reactions
from .fix_get_gpr_association import get_gpr_associations
from .fix_get_highest_compound_type import get_highest_compound_type
from .fix_get_kinetic_data import get_kinetic_data
from .fix_set_reaction_stoichiometry import fix_set_reaction_stoichiometry
from .fix_unify_reaction_direction import fix_unify_reaction_direction



from __future__ import annotations

import logging
from typing import Dict

from ..data import Reaction

logger = logging.getLogger(__name__)


def _reverse_stoichiometry(rxn: Reaction) -> None:
    """Reverse the stoichiometry of a reaction.

    This also reverses the compartments and the gibbs0.
    """
    substrates = rxn.substrates.copy()
    products = rxn.products.copy()
    rxn.substrates = {k: -v for k, v in products.items()}
    rxn.products = {k: -v for k, v in substrates.items()}
    if rxn.gibbs0 is not None:
        rxn.gibbs0 = -rxn.gibbs0
    rxn.substrate_compartments, rxn.product_compartments = (
        rxn.product_compartments,
        rxn.substrate_compartments,
    )


def fix_unify_reaction_direction(parse_reactions: Dict[str, Reaction]) -> Dict[str, Reaction]:
    """Set every reaction to be LEFT-TO-RIGHT and add bounds accordingly."""
    for reaction in parse_reactions.values():
        if reaction.reversible:
            reaction.bounds = (-1000, 1000)
        else:
            direction = reaction.direction
            if direction in (
                "LEFT-TO-RIGHT",
                "PHYSIOL-LEFT-TO-RIGHT",
                "IRREVERSIBLE-LEFT-TO-RIGHT",
            ):
                reaction.bounds = (0, 1000)
            elif direction in (
                "RIGHT-TO-LEFT",
                "PHYSIOL-RIGHT-TO-LEFT",
                "IRREVERSIBLE-RIGHT-TO-LEFT",
            ):
                _reverse_stoichiometry(reaction)
                reaction.bounds = (0, 1000)
            else:
                logger.info(
                    f"Weird reaction direction '{direction}' for reaction {reaction.id}, setting to LEFT-TO-RIGHT"
                )
                reaction.bounds = (0, 1000)
    return parse_reactions
